#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
//----------------------------------------------------------------------
#define MAX_LINES   20
#define MAX_COLUMNS 80
//----------------------------------------------------------------------
static char *man_page_directory = "/usr/share/man";
//----------------------------------------------------------------------
static void  usage          (void);
static FILE *open_man_page  (int section, char *name);
static void build_path_name (int section, char *name, char *path);
//----------------------------------------------------------------------
int
main (int argc, char *argv[], char *envp[])
{
    FILE       *fp;
    char       *name;
    int         section;
    int         c;
    int         line   = 1; // Line internal counter.
    int         column = 1; // Column internal counter.
    int         loop;
    //
    // There must be minimum an argument, and maximum two.
    //
    if (argc < 2 || argc > 3)
      {
        usage ();
        return (1);
      }
    //
    // If there are two arguments, there must be the
    // section number.
    //
    if (argc == 3)
      {
        section = atoi (argv[1]);
        name = argv[2];
      }
    else
      {
        section = 0;
        name = argv[1];
      }
    //
    // Try to open the manual page.
    //
    fp = open_man_page (section, name);
    //
    if (fp == NULL)
      {
        //
        // Error opening file.
        //
        return (1);
      }

    //
    // The following loop continues while the file
    // gives characters, or when a command to change
    // file or to quit is given.
    //
    for (loop = 1; loop; )
      {
        //
        // Read a single character.
        //
        c = getc (fp);
        //
        if (c == EOF)
          {
            loop = 0;
            break;
          }
        //
        // If the character read is a special one,
        // the line/column calculation is modified,
        // so that it is known when to stop scrolling.
        //
        switch (c)
          {
            case '\r':
                //
                // Displaying this character, the cursor should go
                // back to the first column. So the column counter
                // is reset.
                //
                column = 1;
                break;
            case '\n':
                //
                // Displaying this character, the cursor should go
                // back to the next line, at the first column.
                // So the column counter is reset and the line
                // counter is incremented.
                //
                line++;
                column = 1;
                break;
            case '\b':
                //
                // Displaying this character, the cursor should go
                // back one position, unless it is already at the
                // beginning.
                //
                if (column > 1)
                  {
                    column--;
                  }
                break;
            default:
                //
                // Any other character must increase the column
                // counter.
                //
                column++;
          }
        //
        // Display the character, even if it is a special one:
        // it is responsibility of the screen device management
        // to do something good with special characters.
        //
        putchar (c);
        //
        // If the column counter is gone beyond the screen columns,
        // then adjust the column counter and increment the line
        // counter.
        //
        if (column > MAX_COLUMNS)
          {
            column -= MAX_COLUMNS;
            line++;
          }
        //
        // Check if there is space for scrolling.
        //
        if (line < MAX_LINES)
          {
            continue;
          }
        //
        // Here, displayed lines are MAX_LINES.
        //
        if (column > 1)
          {
            //
            // Something was printed at the current line: must
            // do a new line.
            //
            putchar ('\n');
          }
        //
        // Show the more prompt.
        //
        printf ("--More--");
        fflush (stdout);
        //
        // Read a character from standard input.
        //
        c = getchar ();
        //
        // Consider command `q', but any other character
        // can be introduced, to let show the next page.
        //
        switch (c)
          {
            case 'Q':
            case 'q':
                //
                // Quit. But must erase the `--More--' prompt.
                //
                printf ("\b \b\b \b\b \b\b \b\b \b");
                printf ("\b \b\b \b\b \b\b \b");
                fclose (fp);
                return (0);
          }
        //
        // Backspace to overwrite `--More--' and the character
        // pressed.
        //
        printf ("\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b\b \b");
        //
        // Reset line/column counters.
        //
        column = 1;
        line = 1;
      }
    //
    // Close the file pointer if it is still open.
    //
    if (fp != NULL)
      {
        fclose (fp);
      }
    //
    return (0);
}
//----------------------------------------------------------------------
static void
usage (void)
{
    fprintf (stderr, "Usage: man [SECTION] NAME\n");
}
//----------------------------------------------------------------------
FILE *
open_man_page (int section, char *name)
{
    FILE       *fp;
    char        path[PATH_MAX];
    struct stat file_status;
    //
    //
    //
    if (section > 0)
      {
        build_path_name (section, name, path);
        //
        // Check if file exists.
        //
        if (stat (path, &file_status) != 0)
          {
            fprintf (stderr, "Man page %s(%i) does not exist!\n",
                     name, section);
            return (NULL);
          }
      }
    else
      {
        //
        // Must try a section.
        //
        for (section = 1; section < 9; section++)
          {
            build_path_name (section, name, path);
            //
            // Check if file exists.
            //
            if (stat (path, &file_status) == 0)
              {
                //
                // Found.
                //
                break;
              }
          }
      }
    //
    // Check if a file was found.
    //
    if (section < 9)
      {
        fp = fopen (path, "r");
        //
        if (fp == NULL)
          {
            //
            // Error opening file.
            //
            perror (path);
            return (NULL);
          }
        else
          {
            //
            // Opened right.
            //
            return (fp);
          }
      }
    else
      {
        fprintf (stderr, "Man page %s does not exist!\n",
                 name);
        return (NULL);
      }
}
//----------------------------------------------------------------------
void
build_path_name (int section, char *name, char *path)
{
    char  string_section[10];
    //
    // Convert the section number into a string.
    //
    sprintf (string_section, "%i", section);
    //
    // Prepare the path to the man file.
    //
    path[0] = 0;
    strcat (path, man_page_directory);
    strcat (path, "/");
    strcat (path, name);
    strcat (path, ".");
    strcat (path, string_section);
}
