#include <kernel/proc.h>
#include <kernel/k_libc.h>
#include <errno.h>
#include <fcntl.h>
//----------------------------------------------------------------------
ssize_t
fd_write (pid_t pid, int fdn, const void *buffer, size_t count)
{
    proc_t  *ps;
    fd_t    *fd;
    ssize_t  size_written;
    //
    // Get process.
    //
    ps = proc_reference (pid);
    //
    // Get file descriptor.
    //
    fd = fd_reference (pid, &fdn);
    if (fd              == NULL   ||
        fd->file        == NULL   ||
        fd->file->inode == NULL )
      {
        //
        // The file descriptor pointer is not valid.
        //
        errset (EBADF);                 // Bad file descriptor.
        return ((ssize_t) -1);
      }
    //
    // Check if it is opened for write.
    //
    if (!(fd->file->oflags & O_WRONLY))
      {
        //
        // The file is not opened for write.
        //
        errset (EINVAL);                // Invalid argument.
        return ((ssize_t) -1);
      }
    //
    // Check if it is a directory.
    //
    if (fd->file->inode->mode & S_IFDIR)
      {
        errset (EISDIR);                // Is a directory.
        return ((ssize_t) -1);
      }
    //
    // It should be a valid type of file to be written. Check if it is
    // opened in append mode: if so, must move the write offset to the
    // end.
    //
    if (fd->fl_flags & O_APPEND)
      {
        fd->file->offset = fd->file->inode->size;
      }
    //
    // Check the kind of file to be written and write it.
    //
    if (fd->file->inode->mode & S_IFBLK ||
        fd->file->inode->mode & S_IFCHR)
      {
        //
        // A device is to be written.
        //
        size_written = dev_io (pid, (dev_t) fd->file->inode->direct[0],
                               DEV_WRITE, fd->file->offset, buffer,
                               count, NULL);
      }
    else if (fd->file->inode->mode & S_IFREG)
      {
        //
        // A regular file is to be written.
        //
        size_written = inode_file_write (fd->file->inode,
                                         fd->file->offset,
                                         buffer, count);
      }
    else
      {
        //
        // Unsupported file type.
        //
        errset (E_FILE_TYPE_UNSUPPORTED);       //File type unsupported.
        return ((ssize_t) -1);
      }
    //
    // Update the file descriptor internal offset.
    //
    if (size_written > 0)
      {
        fd->file->offset += size_written;
      }
    //
    // Just return the size written, even if it is an error.
    //
    return (size_written);
}
