#include <kernel/fs.h>
#include <errno.h>
#include <kernel/proc.h>
#include <libgen.h>
#include <kernel/k_libc.h>
//----------------------------------------------------------------------
int
path_unlink (pid_t pid, const char *path)
{
    proc_t      *ps;
    inode_t     *inode_unlink;
    inode_t     *inode_directory;
    char         path_unlink[PATH_MAX];
    char         path_copy[PATH_MAX];
    char        *path_directory;
    char        *name_unlink;
    dev_t        device;
    off_t        start;
    char         buffer[SB_MAX_ZONE_SIZE];
    directory_t *dir = (directory_t *) buffer;
    int          status;
    ssize_t      size_read;
    ssize_t      size_written;
    int          d;                     // Directory buffer index.
    //
    // Get process.
    //
    ps = proc_reference (pid);
    //
    // Get full paths.
    //
    path_full (path, ps->path_cwd, path_unlink);
    strncpy (path_copy, path_unlink, PATH_MAX);
    path_directory = dirname (path_copy);
    //
    // Get the inode to be unlinked.
    //
    inode_unlink = path_inode (pid, path_unlink);
    if (inode_unlink == NULL)
      {
        return (-1);
      }
    //
    // If it is a directory, verify that it is empty.
    //
    if (S_ISDIR (inode_unlink->mode))
      {
        if (!inode_dir_empty (inode_unlink))
          {
            inode_put (inode_unlink);
            errset (ENOTEMPTY);         // Directory not empty.
            return (-1);
          }
      }
    //
    // Get the inode of the directory containing it.
    //
    inode_directory = path_inode (pid, path_directory);
    if (inode_directory == NULL)
      {
        inode_put (inode_unlink);
        return (-1);
      }
    //
    // Check if something is mounted on the directory.
    //
    if (inode_directory->sb_attached != NULL)
      {
        //
        // Must select the right directory.
        //
        device = inode_directory->sb_attached->device;
        inode_put (inode_directory);
        inode_directory = inode_get (device, 1);
        if (inode_directory == NULL)
          {
            inode_put (inode_unlink);
            return (-1);
          }
      }
    //
    // Check if write is allowed for the file system.
    //
    if (inode_directory->sb->options & MOUNT_RO)
      {
        errset (EROFS);         // Read-only file system.
        return (-1);
      }
    //
    // Verify access permissions for the directory. The number "3" means
    // that the user must have access permission and write permission:
    // "-wx" == 2+1 == 3.
    //
    status = inode_check (inode_directory, S_IFDIR, 3, ps->uid);
    if (status != 0)
      {
        errset (EPERM);                 // Operation not permitted.
        inode_put (inode_unlink);
        inode_put (inode_directory);
        return (-1);
      }
    //
    // Get the base name to be unlinked: this will alter the
    // original path.
    //
    name_unlink = basename (path_unlink);
    //
    // Read the directory content and try to locate the item to unlink.
    //
    for (start  = 0;
         start <  inode_directory->size;
         start += inode_directory->sb->blksize)
      {
        size_read = inode_file_read (inode_directory, start, buffer,
                                     inode_directory->sb->blksize,
                                     NULL);
        if (size_read < sizeof (directory_t))
          {
            break;
          }
        //
        // Scan the directory portion just read, for the item to unlink.
        //
        dir = (directory_t *) buffer;
        //
        for (d = 0; d < size_read; d += (sizeof (directory_t)), dir++)
          {
            if (dir->ino != 0                                    &&
                strncmp (dir->name, name_unlink, NAME_MAX) == 0)
              {
                //
                // Found the corresponding item: unlink the inode.
                //
                dir->ino     = 0;
                //
                // Update the directory inside the file system.
                //
                size_written = inode_file_write (inode_directory, start,
                                                 buffer, size_read);
                if (size_written != size_read)
                  {
                    //
                    // Write problem: just tell.
                    //
                    k_printf ("kernel alert: directory write error!\n");
                  }
                //
                // Update directory inode and put inode. If the unlinked
                // inode was a directory, the parent directory inode
                // must reduce the file system link count.
                //
                if (S_ISDIR (inode_unlink->mode))
                  {
                    inode_directory->links--;
                  }
                inode_directory->time = k_time (NULL);
                inode_directory->changed = 1;
                inode_put (inode_directory);
                //
                // Reduce link inside unlinked inode and put inode.
                //
                inode_unlink->links--;
                inode_unlink->changed = 1;
                inode_unlink->time = k_time (NULL);
                inode_put (inode_unlink);
                //
                // Just return, as the work is done.
                //
                return (0);
              }
          }
      }
    //
    // At this point, it was not possible to unlink the file.
    //
    inode_put (inode_unlink);
    inode_put (inode_directory);
    errset (EUNKNOWN);          // Unknown error.
    return (-1);
}
