#ifndef _KERNEL_PROC_H
#define _KERNEL_PROC_H  1

#include <kernel/devices.h>
#include <kernel/memory.h>
#include <kernel/fs.h>
#include <kernel/tty.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/os16.h>
#include <stddef.h>
#include <stdint.h>
#include <time.h>

//----------------------------------------------------------------------
#define CLOCK_FREQUENCY_DIVISOR     65535 // [1]
//
// [1]
// Internal clock frequency is (3579545/3) Hz.
// This value is divided by 65535 (0xFFFF) giving 18.2 Hz.
// The divisor value, 65535, if fixed!
//
//----------------------------------------------------------------------
#define PROC_EMPTY                      0
#define PROC_CREATED                    1
#define PROC_READY                      2
#define PROC_RUNNING                    3
#define PROC_SLEEPING                   4
#define PROC_ZOMBIE                     5
//----------------------------------------------------------------------
#define MAGIC_OS16             0x6F733136L  // os16
#define MAGIC_OS16_APPL        0x6170706CL  // appl
#define MAGIC_OS16_KERN        0x6B65726EL  // kern
//----------------------------------------------------------------------
#define PROCESS_MAX  16                 // Process slots.

typedef struct {
    pid_t             ppid;             // Parent PID.
    pid_t             pgrp;             // Process group ID.
    uid_t             uid;              // Real user ID
    uid_t             euid;             // Effective user ID.
    uid_t             suid;             // Saved user ID.
    dev_t             device_tty;	// Controlling terminal.
    char              path_cwd[PATH_MAX];
                                        // Working directory path.
    inode_t          *inode_cwd;	// Working directory inode.
    int               umask;            // File creation mask.
    unsigned long int sig_status;       // Active signals.
    unsigned long int sig_ignore;       // Signals to be ignored.
    clock_t           usage;            // Clock ticks CPU time usage.
    unsigned int      status;
    int               wakeup_events;    // Wake up for something.
    int               wakeup_signal;    // Signal waited.
    unsigned int      wakeup_timer;     // Seconds to wait for.
    addr_t            address_i;
    segment_t         segment_i;
    size_t            size_i;
    addr_t            address_d;
    segment_t         segment_d;
    size_t            size_d;
    uint16_t          sp;
    int               ret;
    char              name[PATH_MAX];
    fd_t              fd[FOPEN_MAX];
} proc_t;

extern proc_t   proc_table[PROCESS_MAX];
//----------------------------------------------------------------------
typedef struct {
    uint32_t filler0;
    uint32_t magic0;
    uint32_t magic1;
    uint16_t segoff;
    uint16_t etext;
    uint16_t edata;
    uint16_t ebss;
    uint16_t ssize;
} header_t;
//----------------------------------------------------------------------
void         _ivt_load          (void);
#define      ivt_load()         (_ivt_load ())
void         proc_init          (void);
void         proc_scheduler     (uint16_t *sp, segment_t *segment);
void         sysroutine         (uint16_t *sp, segment_t *segment,
                                 uint16_t syscallnr, uint16_t msg_off,
                                 uint16_t msg_size);
proc_t      *proc_reference     (pid_t pid);
//----------------------------------------------------------------------
int          proc_sys_exec      (uint16_t *sp, segment_t *segment_d,
                                 pid_t pid, const char *path,
                                 unsigned int argc, char *arg_data,
                                 unsigned int envc, char *env_data);
void         proc_sys_exit      (pid_t pid, int status);
pid_t        proc_sys_fork      (pid_t ppid, uint16_t sp);
int          proc_sys_kill      (pid_t pid_killer, pid_t pid_target,
                                 int sig);
int          proc_sys_seteuid   (pid_t pid, uid_t euid);
int          proc_sys_setuid    (pid_t pid, uid_t uid);
sighandler_t proc_sys_signal    (pid_t pid, int sig,
                                 sighandler_t handler);
pid_t        proc_sys_wait      (pid_t pid, int *status);
//----------------------------------------------------------------------
void         proc_dump_memory   (pid_t pid, addr_t address,
                                 size_t size, char *name);
void         proc_available     (pid_t pid);
pid_t        proc_find          (segment_t segment_d);
void         proc_sch_signals   (void);
void         proc_sch_terminals (void);
void         proc_sch_timers    (void);
void         proc_sig_chld      (pid_t parent, int sig);
void         proc_sig_cont      (pid_t pid, int sig);
void         proc_sig_core      (pid_t pid, int sig);
int          proc_sig_ignore    (pid_t pid, int sig);
void         proc_sig_off       (pid_t pid, int sig);
void         proc_sig_on        (pid_t pid, int sig);
int          proc_sig_status    (pid_t pid, int sig);
void         proc_sig_stop      (pid_t pid, int sig);
void         proc_sig_term      (pid_t pid, int sig);

#endif
