#include <kernel/proc.h>
#include <kernel/k_libc.h>
#include <string.h>
//----------------------------------------------------------------------
extern uint16_t _etext;
//----------------------------------------------------------------------
void
proc_init (void)
{
    uint8_t  divisor_lo;
    uint8_t  divisor_hi;
    pid_t    pid;
    int      fdn;                // File descriptor index;
    addr_t   start;              // Used for effective memory addresses.
    size_t   size;               // Used for memory allocation.
    inode_t *inode;
    sb_t    *sb;
    //
    // Clear interrupts (should already be cleared).
    //
    cli ();
    //
    // Load Interrupt vector table (IVT).
    //
    ivt_load ();
    //
    // Configure the clock: must be the original values, because
    // the BIOS depends on it!
    //
    // Base frequency is 1193181 Hz and it should divided.
    // Resulting frequency must be from 18.22 Hz and 1193181 Hz.
    // The calculated value (the divisor) must be sent to the
    // PIT (programmable interval timer), divided in two pieces.
    //
    divisor_lo = (CLOCK_FREQUENCY_DIVISOR & 0xFF);         // Low  byte.
    divisor_hi = (CLOCK_FREQUENCY_DIVISOR / 0x100) & 0xFF; // High byte.
    out_8 (0x43, 0x36);
    out_8 (0x40, divisor_lo);        // Lower byte.
    out_8 (0x40, divisor_hi);        // Higher byte.
    //
    // Set all memory reference to some invalid data.
    //
    for (pid = 0; pid < PROCESS_MAX; pid++)
      {
        proc_available (pid);
      }
    //
    // Mount root file system.
    //
    inode = NULL;
    sb = sb_mount (DEV_DSK0, &inode, MOUNT_DEFAULT);
    if (sb == NULL || inode == NULL)
      {
        k_perror ("Kernel panic: cannot mount root file system:");
        k_exit (0);
      }
    //
    // Set up the process table with the kernel.
    //
    proc_table[0].ppid          = 0;
    proc_table[0].pgrp          = 0;
    proc_table[0].uid           = 0;
    proc_table[0].euid          = 0;
    proc_table[0].suid          = 0;
    proc_table[0].device_tty    = DEV_UNDEFINED;
    proc_table[0].sig_status    = 0;
    proc_table[0].sig_ignore    = 0;
    proc_table[0].usage         = 0;
    proc_table[0].status        = PROC_RUNNING;
    proc_table[0].wakeup_events = 0;
    proc_table[0].wakeup_signal = 0;
    proc_table[0].wakeup_timer  = 0;
    proc_table[0].segment_i     = seg_i ();
    proc_table[0].address_i     = seg_i ();
    proc_table[0].address_i    *= 16;
    proc_table[0].size_i        = (size_t) &_etext;
    proc_table[0].segment_d     = seg_d ();
    proc_table[0].address_d     = seg_d ();
    proc_table[0].address_d    *= 16;
    proc_table[0].size_d        = 0;     // Maximum size: 0x10000.
    proc_table[0].sp            = 0;     // To be set at next interrupt.
    proc_table[0].ret           = 0;
    proc_table[0].umask         = 0022;  // Default umask.
    proc_table[0].inode_cwd     = inode; // Root fs inode.
    strncpy (proc_table[0].path_cwd, "/", PATH_MAX);
    strncpy (proc_table[0].name, "os16 kernel", PATH_MAX);
    //
    // Ensure to have a terminated string.
    //
    proc_table[0].name[PATH_MAX-1] = 0;
    //
    // Reset file descriptors.
    //
    for (fdn = 0; fdn < OPEN_MAX; fdn++)
      {
        proc_table[0].fd[fdn].fl_flags = 0;
        proc_table[0].fd[fdn].fd_flags = 0;
        proc_table[0].fd[fdn].file     = NULL;
      }
    //
    // Allocate memory for the code segment.
    //
    mb_alloc (proc_table[0].address_i, proc_table[0].size_i);
    //
    // Allocate memory for the data segment if different.
    //
    if (seg_d () != seg_i ())
      {
        mb_alloc (proc_table[0].address_d, proc_table[0].size_d);
      }
    //
    // Allocate memory for the BIOS data area (BDA).
    //
    mb_alloc (0x00000L, 0x500);
    //
    // Allocate memory for the extra BIOS at the
    // bottom of the 640 Kibyte.
    //
    start  = int12 ();
    start *= 1024;
    size   = 0xA0000L - start;
    mb_alloc (start, size);
    //
    // Enable and disable hardware interrupts (IRQ).
    //
    irq_on  (0); // timer.
    irq_on  (1); // enable keyboard
    irq_off (2); // 
    irq_off (3); // 
    irq_off (4); // 
    irq_off (5); // 
    irq_on  (6); // floppy (must be on to let int 13 work)!
    irq_off (7); //
    //
    // Interrupts activation.
    //
    sti ();
}
