#include <kernel/proc.h>
#include <errno.h>
//----------------------------------------------------------------------
pid_t
proc_sys_fork (pid_t ppid, uint16_t sp)
{
    pid_t     pid;
    pid_t     zombie;
    memory_t  allocated_i;
    memory_t  allocated_d;
    int       status;
    int       fdn;
    //
    // Find a free PID.
    //
    for (pid = 1; pid < PROCESS_MAX; pid++)
      {
        if (proc_table[pid].status == PROC_EMPTY)
          {
            break;
          }
      }
    if (pid >= PROCESS_MAX)
      {
        //
        // There is no free pid.
        //
        errset (ENOMEM);        // Not enough space.
        return (-1);
      }
    //
    // Before allocating a new process, must check if there are some
    // zombie slots, still with original segment data: should reset
    // it now!
    //
    for (zombie = 1; zombie < PROCESS_MAX; zombie++)
      {
        if (   proc_table[zombie].status == PROC_ZOMBIE
            && proc_table[zombie].segment_d != -1)
          {
            proc_table[zombie].segment_i = -1;    // Reset
            proc_table[zombie].address_i = -1L;   // memory
            proc_table[zombie].size_i    =  0;    // allocation
            proc_table[zombie].segment_d = -1;    // data
            proc_table[zombie].address_d = -1L;   // to
            proc_table[zombie].size_d    =  0;    // impossible
            proc_table[zombie].sp        =  0;    // values.
          }
      }
    //
    // Allocate memory: code and data segments.
    //
    if (proc_table[ppid].segment_i == proc_table[ppid].segment_d)
      {
        //
        // Code segment and Data segment are the same
        // (same I&D).
        //
        status = mb_alloc_size (proc_table[ppid].size_i, &allocated_i);
        if (status < 0)
          {
            errset (ENOMEM);        // Not enough space.
            return ((pid_t) -1);
          }
        allocated_d.address = allocated_i.address;
        allocated_d.segment = allocated_i.segment;
        allocated_d.size    = allocated_i.size;
      }
    else
      {
        //
        // Code segment and Data segment are different
        // (different I&D).
        // Only the data segment is allocated.
        //
        status = mb_alloc_size (proc_table[ppid].size_d, &allocated_d);
        if (status < 0)
          {
            errset (ENOMEM);        // Not enough space.
            return ((pid_t) -1);
          }
        //
        // Code segment is the same from the parent process.
        //
        allocated_i.address = proc_table[ppid].address_i;
        allocated_i.segment = proc_table[ppid].segment_i;
        allocated_i.size    = proc_table[ppid].size_i;
      }
    //
    // Copy the process in memory.
    //
    if (proc_table[ppid].segment_i == proc_table[ppid].segment_d)
      {
        //
        // Code segment and data segment are the same:
        // must copy all.
        //
        // Copy the code segment: if the size is zero,
        // it means 0x10000 bytes (65536 bytes).
        //
        if (proc_table[ppid].size_i == 0)
          {
            //
            // Copy 0x10000 bytes with two steps.
            //
            mem_copy (proc_table[ppid].address_i,
                      allocated_i.address, 0x8000);
            mem_copy ((proc_table[ppid].address_i + 0x8000),
                      (allocated_i.address + 0x8000), 0x8000);
          }
        else
          {
            //
            // Normal copy.
            //
            mem_copy (proc_table[ppid].address_i, allocated_i.address,
                      proc_table[ppid].size_i);
          }
      }
    else
      {
        //
        // Code segment and data segment are different:
        // copy only the data segment.
        //
        // Copy the data segment in memory: if the size is zero,
        // it means 0x10000 bytes (65536 bytes).
        //
        if (proc_table[ppid].size_d == 0)
          {
            //
            // Copy 0x10000 bytes with two steps.
            //
            mem_copy (proc_table[ppid].address_d,
                      allocated_d.address, 0x8000);
            mem_copy ((proc_table[ppid].address_d + 0x8000),
                      (allocated_d.address + 0x8000), 0x8000);
          }
        else
          {
            //
            // Normal copy.
            //
            mem_copy (proc_table[ppid].address_d, allocated_d.address,
                      proc_table[ppid].size_d);
          }
      }
    //
    // Allocate the new PID.
    //
    proc_table[pid].ppid          = ppid;
    proc_table[pid].pgrp          = proc_table[ppid].pgrp;
    proc_table[pid].uid           = proc_table[ppid].uid;
    proc_table[pid].euid          = proc_table[ppid].euid;
    proc_table[pid].suid          = proc_table[ppid].suid;
    proc_table[pid].device_tty    = proc_table[ppid].device_tty;
    proc_table[pid].sig_status    = 0;
    proc_table[pid].sig_ignore    = 0;
    proc_table[pid].usage         = 0;
    proc_table[pid].status        = PROC_CREATED;
    proc_table[pid].wakeup_events = 0;
    proc_table[pid].wakeup_signal = 0;
    proc_table[pid].wakeup_timer  = 0;
    proc_table[pid].segment_i     = allocated_i.segment;
    proc_table[pid].address_i     = allocated_i.address;
    proc_table[pid].size_i        = proc_table[ppid].size_i;
    proc_table[pid].segment_d     = allocated_d.segment;
    proc_table[pid].address_d     = allocated_d.address;
    proc_table[pid].size_d        = proc_table[ppid].size_d;
    proc_table[pid].sp            = sp;
    proc_table[pid].ret           = 0;
    proc_table[pid].inode_cwd     = proc_table[ppid].inode_cwd;
    proc_table[pid].umask         = proc_table[ppid].umask;
    strncpy (proc_table[pid].name,
             proc_table[ppid].name, PATH_MAX);
    strncpy (proc_table[pid].path_cwd,
             proc_table[ppid].path_cwd, PATH_MAX);
    //
    // Increase inode references for the working directory.
    //
    proc_table[pid].inode_cwd->references++;
    //
    // Duplicate valid file descriptors.
    //
    for (fdn = 0; fdn < OPEN_MAX; fdn++)
      {
        if (   proc_table[ppid].fd[fdn].file != NULL
            && proc_table[ppid].fd[fdn].file->inode != NULL)
          {
            //
            // Copy to the forked process.
            //
            proc_table[pid].fd[fdn].fl_flags
                = proc_table[ppid].fd[fdn].fl_flags;
            proc_table[pid].fd[fdn].fd_flags
                = proc_table[ppid].fd[fdn].fd_flags;
            proc_table[pid].fd[fdn].file
                = proc_table[ppid].fd[fdn].file;
            //
            // Increment file reference.
            //
            proc_table[ppid].fd[fdn].file->references++;
          }
      }
    //
    // Change segment values inside the stack: DS==ES; CS.
    //
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + proc_table[pid].sp + 14),
            &allocated_d.segment, (sizeof allocated_d.segment), NULL);
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + proc_table[pid].sp + 16),
            &allocated_d.segment, (sizeof allocated_d.segment), NULL);
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + proc_table[pid].sp + 20),
            &allocated_i.segment, (sizeof allocated_i.segment), NULL);
    //
    // Set it ready.
    //
    proc_table[pid].status  = PROC_READY;
    //
    // Return the new PID.
    //
    return (pid);
}
